//
//    Title: imagegray.ccp
//  Created: Thu Oct  13 15:53:57 2005
//   Author: Denis Steinemann <deniss@inf.ethz.ch> (based on earlier Version by Tim Weyrich)
//
//  copyright (c) 2001--2005, Computer Graphics Laboratory, ETH Zuerich
//

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <iostream>
#include <math.h>

#include "imagegray.h"
#include "clip.h"

#include <tiff.h>
#include <tiffio.h>

#define  HAVE_BIGENDIAN  (*(unsigned short *)"12" == 0x3132)

#define  SWAP_UINT32(x)  (( ((unsigned long)(x)) << 24) |  ((((unsigned long)(x)) & 0x0000ff00UL) <<  8) |  ((((unsigned long)(x)) & 0x00ff0000UL) >>  8) |  ( ((unsigned long)(x))                 >> 24))

///////////////////////////////////////////////////////////////////////////////

ImageGray  *imageGrayNew(int width, int height)
  /*! After use, the image must be deleted by imageGrayFree().
   *
   *    \param width            width of the new image in pixels
   *    \param height   height of the new image in pixels
   *
   *    \return         a pointer to the new image struct
   */
{
  ImageGray  *result = NULL;
  
  assert(width >= 0);
  assert(height >= 0);

  result = (ImageGray *)malloc(sizeof(ImageGray));
  result->width = width;
  result->height = height;
  result->bytesPerPixel = 1;
  result->data = (unsigned char *)malloc(width * height * sizeof(unsigned char));

  return result;
}

void  imageGrayFree(ImageGray *img)
  /*! The memory used by the image is freed. All pointers to the
   *  image become invalid. After use, the created image must be
   *  deleted by imageGrayFree().
   *
   *    \param img      the image to be deleted
   */
{
  assert(img->bytesPerPixel == 1);
  
  free(img->data);
  free(img);

  img = NULL;
}

// ----------------------------------------------------------------------------

ImageGray  *imageGrayClone(const ImageGray *img)
  /*! The copy has got the same extent and image data.
   *
   *    \param img      source image
   *
   *    \return a pointer to the new ImageGray struct
   */
{
  assert(img->bytesPerPixel == 1);
  
  ImageGray  *result = imageGrayNew(img->width, img->height);

  if (result)
    memcpy(result->data, img->data, img->width * img->height * sizeof(unsigned char));
  
  return result;
}


// ----------------------------------------------------------------------------

void  imageGrayCopy(const ImageGray *src, const ImageGray *dest)
  /*! src and dest have got the same extents.
   *
   *    \param src      source image
   *    \param dest     source image
   *
   */
{
  assert(src->height == dest->height && src->width == dest->width);
  
  memcpy(dest->data, src->data, src->width * src->height * sizeof(unsigned char));
 
}

///////////////////////////////////////////////////////////////////////////////

void  imageGrayClear(ImageGray *img)
  /*! All pixels in the image are set to white. So, a call to
   *  imageGrayClear(img) is identical to imageGrayClear(img, 255).
   *
   *    \param img the image to be cleared
   */
{
  assert(img->bytesPerPixel == 1);
  imageGraySet(img, 255);
}

void  imageGraySet(ImageGray *img, unsigned val)
  /*! All pixels in the image are set to \a val.
   *
   *    \param img      the destination image
   */
{
  assert(img->bytesPerPixel == 1);
  memset(img->data, val, img->width * img->height);
}

///////////////////////////////////////////////////////////////////////////////

#define  RGB2GRAY_QUOTIENT      65536

#define  RGB2GRAY_RED_WEIGHT    (RGB2GRAY_QUOTIENT*30/100)
#define  RGB2GRAY_GREEN_WEIGHT  (RGB2GRAY_QUOTIENT*59/100)
#define  RGB2GRAY_BLUE_WEIGHT   (RGB2GRAY_QUOTIENT*11/100)

ImageGray  *imageGrayReadTIFF(const char *fname)
  /*! The image is read from a file named \a fname. If the file
   *  doesn't exist, or the file is no valid TIFF file, \e NULL is
   *  returned.
   *
   *  If the image at \a fname is a color image, it is converted
   *  to a gray-level image automatically.
   *
   *    \param fname            filename
   *
   *    \return the read image; \e NULL on error
   */
{
  TIFF  *tif;
  if ((tif = TIFFOpen(fname, "r")) != NULL)
    {
      ImageGray  *result = NULL;
      uint32      width, height;

      TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &width);
      TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &height);
      
      if ((result = imageGrayNew(width, height)) != NULL)
        {
          unsigned long *buf = (unsigned long*)malloc(result->width * result->height
                                                        * sizeof(unsigned long));
          // I'm very lazy, so I'm reading the image as rgba-image and
          // convert it afterwards...
	  if (TIFFReadRGBAImage(tif, result->width, result->height, ((uint32*)buf), 0))
            { // success -- flip image
              unsigned long  *p;
              unsigned char  *q;
              int             c;
              
              for (p=buf+result->width*(result->height-1),
                     q=result->data;
                   p >= buf;
                   p-=result->width, q+=result->width)
                for (c=result->width-1; c>=0; c--)
                  q[c] = (unsigned char)((RGB2GRAY_RED_WEIGHT   * TIFFGetR(p[c]) +
                                          RGB2GRAY_GREEN_WEIGHT * TIFFGetG(p[c]) +
                                          RGB2GRAY_BLUE_WEIGHT  * TIFFGetB(p[c]))
                                         / RGB2GRAY_QUOTIENT);
            }
          else
            {
              imageGrayFree(result);
              result = NULL;
            }
        }

      TIFFClose(tif);
      return result;
    }
  
  return NULL;
}

int  imageGrayWriteTIFF(const char *fname, const ImageGray *img)
  /*! The image is written to a file named \a fname. If the file
   *  doesn't exist, a new one is created. Otherwise, the existing
   *  file is overwritten. If an error occurs, an error value !=0
   *  is returned.
   *
   *    \param fname            filename
   *
   *    \return         0 on success
   */
{
  assert(img->bytesPerPixel == 1);

  TIFF  *tif;
  if ((tif = TIFFOpen(fname, "w")) != NULL)
    {
      int  y;

      TIFFSetField(tif, TIFFTAG_SUBFILETYPE, 0);
      TIFFSetField(tif, TIFFTAG_IMAGEWIDTH, img->width);
      TIFFSetField(tif, TIFFTAG_IMAGELENGTH, img->height);
      TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, 1);
      TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE, 8);
      TIFFSetField(tif, TIFFTAG_COMPRESSION, COMPRESSION_PACKBITS);
      TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK);
      TIFFSetField(tif, TIFFTAG_DOCUMENTNAME, fname);
      TIFFSetField(tif, TIFFTAG_IMAGEDESCRIPTION, "Created With My Personal Painter Application :-)");
      TIFFSetField(tif, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
      TIFFSetField(tif, TIFFTAG_ROWSPERSTRIP, 64);
      TIFFSetField(tif, TIFFTAG_XRESOLUTION, 75.0);
      TIFFSetField(tif, TIFFTAG_YRESOLUTION, 75.0);
      TIFFSetField(tif, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
      TIFFSetField(tif, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);

      for (y=0; y<img->height; y++)
        if (TIFFWriteScanline(tif, img->data+y*img->width, y, 0) < 0)
          break;

      if (y < img->height)
        {
          //cerr << "Error while writing to TIFF file." << endl;
          TIFFClose(tif);
          return 1;
        }
      else
        {
          TIFFClose(tif);
          return 0;
        }
    }
  else
    //cerr << "Couldn't open `" << fname << "' for writing." << endl;
  
  return 1;
}

///////////////////////////////////////////////////////////////////////////////

void  imageGrayDrawLine(ImageGray *img, int x0, int y0, int x1, int y1, unsigned char grayValue)
{
  // clip line
  {
    float  x[2] = {(float)x0, (float)x1};
    float  y[2] = {(float)y0, (float)y1};

    if (clip(x, y, 0, 0, (float)img->width-1, (float)img->height-1) == 2)
      return;

    x0 = (int)floor(x[0] + 0.5);
    y0 = (int)floor(y[0] + 0.5);
    x1 = (int)floor(x[1] + 0.5);
    y1 = (int)floor(y[1] + 0.5);
  }
  
  // starting point of line
  int x = x0, y = y0;

  // direction of line
  int dx = x1-x0, dy = y1-y0;

  // increment or decrement depending on direction of line
  int sx = (dx > 0 ? 1 : (dx < 0 ? -1 : 0));
  int sy = (dy > 0 ? 1 : (dy < 0 ? -1 : 0));

  // decision parameters for voxel selection
  if ( dx < 0 ) dx = -dx;
  if ( dy < 0 ) dy = -dy;
  int ax = 2*dx, ay = 2*dy;
  int decx, decy;

  // determine largest direction component, single-step related variable
  int max = dx, var = 0;
  if ( dy > max ) { var = 1; }

  // traverse Bresenham line
  switch ( var )
    {
    case 0:  // single-step in x-direction
      for (decy=ay-dx; /**/; x += sx, decy += ay)
        {
          // process pixel
          img->data[x + y*img->width] = grayValue;

          // take Bresenham step
          if ( x == x1 ) break;
          if ( decy >= 0 ) { decy -= ax; y += sy; }
        }
      break;
    case 1:  // single-step in y-direction
      for (decx=ax-dy; /**/; y += sy, decx += ax)
        {
          // process pixel
          img->data[x + y*img->width] = grayValue;

          // take Bresenham step
          if ( y == y1 ) break;
          if ( decx >= 0 ) { decx -= ay; x += sx; }
        }
      break;
    }
}

///////////////////////////////////////////////////////////////////////////////

void  imageGrayInvertLine(ImageGray *img, int x0, int y0, int x1, int y1)
{
  // clip line
  {
    float  x[2] = {(float)x0, (float)x1};
    float  y[2] = {(float)y0, (float)y1};

    if (clip(x, y, 0.0f, 0.0f, (float)(img->width-1), (float)(img->height-1)) == 2)
      return;

    x0 = (int)floor(x[0] + 0.5);
    y0 = (int)floor(y[0] + 0.5);
    x1 = (int)floor(x[1] + 0.5);
    y1 = (int)floor(y[1] + 0.5);
  }
  
  // starting point of line
  int x = x0, y = y0;

  // direction of line
  int dx = x1-x0, dy = y1-y0;

  // increment or decrement depending on direction of line
  int sx = (dx > 0 ? 1 : (dx < 0 ? -1 : 0));
  int sy = (dy > 0 ? 1 : (dy < 0 ? -1 : 0));

  // decision parameters for voxel selection
  if ( dx < 0 ) dx = -dx;
  if ( dy < 0 ) dy = -dy;
  int ax = 2*dx, ay = 2*dy;
  int decx, decy;

  // determine largest direction component, single-step related variable
  int max = dx, var = 0;
  if ( dy > max ) { var = 1; }

  // traverse Bresenham line
  switch ( var )
    {
    case 0:  // single-step in x-direction
      for (decy=ay-dx; /**/; x += sx, decy += ay)
        {
          // process pixel
          img->data[x + y*img->width] = 255 - img->data[x + y*img->width];

          // take Bresenham step
          if ( x == x1 ) break;
          if ( decy >= 0 ) { decy -= ax; y += sy; }
        }
      break;
    case 1:  // single-step in y-direction
      for (decx=ax-dy; /**/; y += sy, decx += ax)
        {
          // process pixel
          img->data[x + y*img->width] = 255 - img->data[x + y*img->width];

          // take Bresenham step
          if ( y == y1 ) break;
          if ( decx >= 0 ) { decx -= ay; x += sx; }
        }
      break;
    }
}

///////////////////////////////////////////////////////////////////////////////

/* draw ellipse incrementally using Bresenham's algorithm */
void  imageGrayDrawEllipse(ImageGray *img, int cx, int cy, int halfaxis_x, int halfaxis_y, unsigned char grayValue)
{
	halfaxis_x = abs(halfaxis_x);

	/* intermediate terms to speed up loop */
	long t1 = halfaxis_x*halfaxis_x, t2 = t1<<1, t3 = t2<<1;
	long t4 = halfaxis_y*halfaxis_y, t5 = t4<<1, t6 = t5<<1;
	long t7 = halfaxis_x*t5, t8 = t7<<1, t9 = 0L;
	long d1 = t2 - t7 + (t4>>1);	/* error terms */
	long d2 = (t1>>1) - t8 + t5;

	int x = halfaxis_x, y = 0;	/* ellipse points */

	while (d2 < 0)			/* til slope = -1 */
	{
		/* draw 4 points using symmetry */
	    imageGrayDrawPoint(img, cx + x, cy + y, grayValue);
	    imageGrayDrawPoint(img, cx + x, cy - y, grayValue);
	    imageGrayDrawPoint(img, cx - x, cy + y, grayValue);
	    imageGrayDrawPoint(img, cx - x, cy - y, grayValue);

		y++;		/* always move up here */
		t9 += t3;	
		if (d1 < 0)	/* move straight up */
		{
			d1 += t9 + t2;
			d2 += t9;
		}
		else		/* move up and left */
		{
			x--;
			t8 -= t6;
			d1 += t9 + t2 - t8;
			d2 += t9 + t5 - t8;
		}
	}

	while (x >= 0) 				/* rest of top right quadrant */
	{
		/* draw 4 points using symmetry */
	    imageGrayDrawPoint(img, cx + x, cy + y, grayValue);
	    imageGrayDrawPoint(img, cx + x, cy - y, grayValue);
	    imageGrayDrawPoint(img, cx - x, cy + y, grayValue);
	    imageGrayDrawPoint(img, cx - x, cy - y, grayValue);

		x--;		/* always move left here */
		t8 -= t6;	
		if (d2 < 0)	/* move up and left */
		{
			y++;
			t9 += t3;
			d2 += t9 + t5 - t8;
		}
		else		/* move straight left */
			d2 += t5 - t8;
	} 
}

///////////////////////////////////////////////////////////////////////////////

/* draw ellipse incrementally using Bresenham's algorithm */
void  imageGrayInvertEllipse(ImageGray *img, int cx, int cy, int halfaxis_x, int halfaxis_y)
{
	halfaxis_x = abs(halfaxis_x);
	
	/* intermediate terms to speed up loop */
	long t1 = halfaxis_x*halfaxis_x, t2 = t1<<1, t3 = t2<<1;
	long t4 = halfaxis_y*halfaxis_y, t5 = t4<<1, t6 = t5<<1;
	long t7 = halfaxis_x*t5, t8 = t7<<1, t9 = 0L;
	long d1 = t2 - t7 + (t4>>1);	/* error terms */
	long d2 = (t1>>1) - t8 + t5;
	unsigned char invGrayValue;

	int x = halfaxis_x, y = 0;	/* ellipse points */

	while (d2 < 0)			/* til slope = -1 */
	{
		/* draw 4 points using symmetry */
		imageGrayDrawInvertedPoint(img, cx + x, cy + y);
	    imageGrayDrawInvertedPoint(img, cx + x, cy - y);
	    imageGrayDrawInvertedPoint(img, cx - x, cy + y);
	    imageGrayDrawInvertedPoint(img, cx - x, cy - y);

		y++;		/* always move up here */
		t9 += t3;	
		if (d1 < 0)	/* move straight up */
		{
			d1 += t9 + t2;
			d2 += t9;
		}
		else		/* move up and left */
		{
			x--;
			t8 -= t6;
			d1 += t9 + t2 - t8;
			d2 += t9 + t5 - t8;
		}
	}

	while (x >= 0) 				/* rest of top right quadrant */
	{
		/* draw 4 points using symmetry */
	    invGrayValue = 255 - img->data[x + y*img->width];

		imageGrayDrawInvertedPoint(img, cx + x, cy + y);
	    imageGrayDrawInvertedPoint(img, cx + x, cy - y);
	    imageGrayDrawInvertedPoint(img, cx - x, cy + y);
	    imageGrayDrawInvertedPoint(img, cx - x, cy - y);

		x--;		/* always move left here */
		t8 -= t6;	
		if (d2 < 0)	/* move up and left */
		{
			y++;
			t9 += t3;
			d2 += t9 + t5 - t8;
		}
		else		/* move straight left */
			d2 += t5 - t8;
	} 
}

///////////////////////////////////////////////////////////////////////////////

inline void  imageGrayDrawPoint(ImageGray *img, int x, int y, unsigned char grayValue)
{
	// process pixel
	if (x >= 0 && y >= 0 && x < img->width && y < img->height) {	
		img->data[x + y*img->width] = grayValue;
	}
}

///////////////////////////////////////////////////////////////////////////////

inline void  imageGrayDrawInvertedPoint(ImageGray *img, int x, int y)
{
	// process pixel
	if (x >= 0 && y >= 0 && x < img->width && y < img->height) {	
		unsigned char invGrayValue = 255 - img->data[x + y*img->width];
		img->data[x + y*img->width] = invGrayValue;
	}
}

