// -*- C++ -*-
//    Title: imagegray.h
//  Created: Thu Oct  13 15:53:57 2005
//   Author: Denis Steinemann <deniss@inf.ethz.ch> (based on earlier Version by Tim Weyrich)
//
//  copyright (c) 2001--2005, Computer Graphics Laboratory, ETH Zuerich
//

#ifndef __IMAGEGRAY_H__
#define __IMAGEGRAY_H__

//! C-struct that holds a gray-level image.
/*! Use imageGrayNew() and imageGrayFree() to create or to delete an
 *  ImageGray struct, respectively.
 */
typedef struct
{
  int    width;                 //!< Width of the image.
  int    height;                //!< Height of the image.

  int    bytesPerPixel;         //!< Number of bytes per Pixel (always 1).
  
  //! Pixel data of the image.
  /*! The data is organized in row-major order. That is, pixels of an
   *  image row are stored subsequently from left to right, and rows
   *  are stored subsequently from top to bottom.
   */
  unsigned char  *data;
} ImageGray;


//! Creates a new gray-level image.
extern ImageGray  *imageGrayNew(int width, int height);

//! Deletes a gray-level image.
extern void        imageGrayFree(ImageGray *img);


//! Creates a copy of a gray-level image.
extern ImageGray  *imageGrayClone(const ImageGray *img);


//! Writes all gray values from a gray-level image \src to a gray-level image \dest. Both images must have the same extensions.
extern void			imageGrayCopy(const ImageGray *src, const ImageGray *dest);


//! Clears a gray-level image.
extern void        imageGrayClear(ImageGray *img);

//! Sets all pixels in the image to a given value.
extern void        imageGraySet(ImageGray *img, unsigned val);

//! Reads a gray-level image from a TIFF file.
extern ImageGray  *imageGrayReadTIFF(const char *fname);

//! Writes a gray-level image as a TIFF file.
extern int         imageGrayWriteTIFF(const char *fname, const ImageGray *img);

//! Draws a line into a gray-level image.
/*! The line is drawn with pixels of gray-value \a grayValue from the
 *  point (\a x0, \a y0) to (\a x1, \a y1).
 */
extern void  imageGrayDrawLine(ImageGray *img, int x0, int y0, int x1, int y1, unsigned char grayValue);

//! Draws a line into a gray-level image by inverting pixels.
/*! The line is drawn from the point (\a x0, \a y0) to (\a x1, \a y1).
 */
extern void  imageGrayInvertLine(ImageGray *img, int x0, int y0, int x1, int y1);


//! Draws an ellipse into a gray-level image.
/*! The ellipse is drawn with pixels of gray-value \a grayValue around the
 *  center point (\a cx, \a cy) with half-axes in x- and y-direction (\a halfaxis_x, \a halfaxis_y).
 */
extern void  imageGrayDrawEllipse(ImageGray *img, int cx, int cy, int halfaxis_x, int halfaxis_y, unsigned char grayValue);


//! Draws an ellipse into a gray-level image by inverting pixels.
/*! The ellipse is drawn around the center point (\a c0, \a c1) to (\a x1, \a y1).
 */
extern void  imageGrayInvertEllipse(ImageGray *img, int cx, int cy, int a, int b);


//! Draws a single-pixel sized point into a gray-level image.
/*! The point is drawn a gray-value \a grayValue at position (\a x, \a y). 
 */
extern inline void  imageGrayDrawPoint(ImageGray *img, int x, int y, unsigned char grayValue);

//! Draws a single-pixel sized point into a gray-level image by inverting the pixel.
/*! The point is drawn at position (\a x, \a y). 
 */
extern inline void  imageGrayDrawInvertedPoint(ImageGray *img, int x, int y);

#endif
